@@ -14,9 +14,9 @@ use itertools::Itertools;
 use crate::{all_imports::RInts, pyo3_prelude::*, sv, util::BitMask128};
 
 use super::{
-    circuit_optimizer::OptimizationContext, computational_node::EinsumArgs, deep_map_op_context,
-    evaluate_fn, prelude::*, py_circuit_items::PY_CIRCUIT_ITEMS, Add, Concat, Einsum,
-    GeneralFunction, HashBytes, Index, Rearrange, ScalarConstant, Scatter, Symbol,
+    circuit_optimizer::OptimizationContext, computational_node::EinsumArgs, evaluate_fn,
+    prelude::*, py_circuit_items::PY_CIRCUIT_ITEMS, Add, Concat, Einsum, GeneralFunction,
+    HashBytes, Index, Rearrange, ScalarConstant, Scatter, Symbol,
 };
 use crate::{
     filter_by_variant,
@@ -109,13 +109,13 @@ pub fn add_deduplicate(add: &Add) -> Option<Add> {
         None
     } else {
         let deduped_inputs = hash_to_node
-            .iter()
+            .into_iter()
             .map(|(hash, node)| {
-                let count = duplicate_counts[hash];
+                let count = duplicate_counts[&hash];
                 if count > 1 {
                     Einsum::scalar_mul(node, count as f64, None).rc()
                 } else {
-                    node.clone()
+                    node
                 }
             })
             .collect();
@@ -1511,3 +1511,91 @@ pub fn extract_add(add: &Add, sub: &Add) -> Option<Add> {
 
     Some(Add::try_from_counts(&counts, add.name_cloned()).unwrap())
 }
+
+/// De-duplicate children of Add nodes that are the same except for a scalar product
+#[pyfunction]
+pub fn add_fuse_scalar_multiples(add: &Add) -> Option<Add> {
+    if !add.children().any(|x| x.as_einsum().is_some()) {
+        return None;
+    }
+
+    let mut reduced_any = 0;
+
+    let scalars_and_children: Vec<(Einsum, f64, Option<String>)> = add
+        .children()
+        .map(|c| match &**c {
+            Circuit::Einsum(e) => {
+                let mut scalar: f64 = 1.0;
+                let mut new_name: Option<String> = None;
+                let no_scalar_children: EinsumArgs = e
+                    .args
+                    .iter()
+                    .filter(|args| {
+                        let (child, _axes) = &**args;
+                        child.as_scalar_constant().map_or(true, |x| {
+                            if x.info().shape.len() == 0 {
+                                scalar *= x.value;
+                                reduced_any += 1;
+                                if new_name.is_none() {
+                                    new_name = x.name_cloned();
+                                }
+                                return false;
+                            }
+                            true
+                        })
+                    })
+                    .map(|x| x.clone())
+                    .collect();
+
+                (
+                    Einsum::try_new(no_scalar_children, e.out_axes.clone(), e.name_cloned())
+                        .unwrap(),
+                    scalar,
+                    new_name,
+                )
+            }
+            Circuit::ScalarConstant(s) => {
+                if s.info().shape.len() == 0 {
+                    (Einsum::empty(None), s.value, s.name_cloned())
+                } else {
+                    (Einsum::identity(c, None), 1.0, None)
+                }
+            }
+            _ => (Einsum::identity(c, None), 1.0, None),
+        })
+        .collect();
+
+    if reduced_any <= 1 {
+        return None;
+    }
+
+    let mut hm: HashMap<Einsum, (f64, Option<String>)> = HashMap::new();
+    for (einsum, scalar, name) in scalars_and_children {
+        let (hm_val, hm_name) = hm.entry(einsum).or_insert((0.0, None));
+        *hm_val += scalar;
+        if hm_name.is_none() {
+            *hm_name = name;
+        }
+    }
+
+    if hm.len() == add.nodes.len() {
+        return None;
+    }
+
+    let nodes: Vec<CircuitRc> = hm
+        .into_iter()
+        .map(|(einsum, (scalar, scalar_name))| {
+            if scalar == 1.0 {
+                einsum.rc()
+            } else {
+                let name = einsum.name_cloned();
+                let out_axes = einsum.out_axes;
+                let mut v = einsum.args;
+                v.push((ScalarConstant::new(scalar, sv![], scalar_name).rc(), sv![]));
+                Einsum::try_new(v, out_axes, name).unwrap().rc()
+            }
+        })
+        .collect();
+
+    return Some(Add::try_new(nodes, add.name_cloned()).unwrap());
+}
