@@ -186,7 +186,7 @@ pub static HASH_TENSOR: GILLazyPy<PyObject> = GILLazyPy::new_py(|py| {
     let module = PyModule::from_code(
         py,
         include_str!(concat!(env!("CARGO_MANIFEST_DIR"), "/tensor_hash.py")),
-        "tensor_hash.py",
+        concat!(env!("CARGO_MANIFEST_DIR"), "/tensor_hash.py"),
         "tensor_hash",
     )
     .unwrap();
